/*
 * Name: mount.c
 * Description: Functions that do the NFS socket setup and NFS mount/unmount.
 *     Parts of it are based on the mount module of p3nfs (Psion NFS).
 * Author: Christian Starkjohann <cs@hal.kph.tuwien.ac.at>
 * Date: 1996-11-14
 * Copyright: GNU-GPL
 * Tabsize: 4
 */
#include <libc.h>

#define NFSCLIENT
#include <rpc/rpc.h>
#include "nfs_prot.h"
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/mount.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <nfs/nfs_mount.h>
#include <mntent.h>
#include "my_defines.h"

#define	DPRINTF(arg)	if(debug_mode & DEBUG_NFS)	printf arg
#define	DDPRINTF(arg)	if(debug_mode & DEBUG_DISPATCH)	printf arg

static char	*mntdir;
static int	do_exit = 0;
static int	do_nfsumount = 0;

#define	MTAB_TMP	(MOUNTED ".tmp")

/* ------------------------------------------------------------------------- */

static void	nfs_unmount(void)
{
FILE			*fpout, *fpin;
struct mntent	*ent;

	fprintf(stderr, "Going to unmount NFS...\n");
	while(unmount(mntdir)){
		perror(mntdir);
		if(errno != EBUSY){
			break;
		}
		sleep(1);
	}
	/* update mtab */
	if((fpout = setmntent(MTAB_TMP, "w")) == NULL){
		perror(MTAB_TMP);
	}else if((fpin = setmntent(MOUNTED, "r")) == NULL){
		endmntent(fpout);
		unlink(MTAB_TMP);
		perror(MOUNTED);
	}else{
		while ((ent = getmntent(fpin)) != NULL){
			if(strcmp(ent->mnt_dir, mntdir) != 0)
				addmntent(fpout, ent);
		}
		endmntent(fpin);
		endmntent(fpout);
		if(rename(MTAB_TMP, MOUNTED) != 0){
			perror(MOUNTED);
			unlink(MTAB_TMP);
		}
	}
	fprintf(stderr, "NFS unmounted.\n");
}

/* ------------------------------------------------------------------------- */

void	terminate(int exitcode)
{
	nfs_unmount();
	exit(exitcode);
}

/* ------------------------------------------------------------------------- */

void	do_unmount(int signr)
{
	do_exit = 1;
	do_nfsumount = 1;
	switch(signr){
	case SIGHUP:	case SIGINT:	case SIGQUIT:
	case SIGUSR1:	case SIGUSR2:
		fprintf(stderr, "** caught signal %d, preparing for unmount\n", signr);
		break;	/* if the signal was user-induced: unmount nfs from mainloop */
	default:
		fprintf(stderr, "** caught signal %d, unmounting\n", signr);
		do_nfsumount = 0;
		nfs_unmount();
	}
}

/* ------------------------------------------------------------------------- */

void	mount_and_dispatch(char *dir, void (*proc)(), nfs_fh *root_fh)
{
int					sock, port, pid, my_pid;
struct sockaddr_in	sain;
SVCXPRT				*nfsxprt;
int					bufsiz = 0xc000;	/* room for a few biods */
struct nfs_args		nfs_args;

	signal(SIGHUP, do_unmount);
	signal(SIGINT, do_unmount);
	signal(SIGQUIT, do_unmount);
	signal(SIGILL, do_unmount);
	signal(SIGIOT, do_unmount);
	signal(SIGTRAP, do_unmount);
	signal(SIGEMT, do_unmount);
	signal(SIGFPE, do_unmount);
	signal(SIGBUS, do_unmount);
	signal(SIGSEGV, do_unmount);
	signal(SIGSYS, do_unmount);
	signal(SIGPIPE, do_unmount);
	signal(SIGTERM, do_unmount);
	signal(SIGUSR1, do_unmount);
	signal(SIGUSR2, do_unmount);
	bzero(&sain, sizeof(struct sockaddr_in));
	bzero(&nfs_args, sizeof(struct nfs_args));

/*** First part: set up the rpc service */
/* Create udp socket */
	sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if(setsockopt(sock, SOL_SOCKET, SO_RCVBUF, (char *)&bufsiz,sizeof(bufsiz)))
		perror("setsockopt");

/* Bind it to a reserved port */
	sain.sin_family = AF_INET;
	sain.sin_addr.s_addr = inet_addr("127.0.0.1");
/*sain.sin_addr.s_addr = htonl(INADDR_ANY);gec*/
	for(port = IPPORT_RESERVED-1; port > IPPORT_RESERVED/2; port--){
	DPRINTF(("trying to use port %d\n", port));
		sain.sin_port = htons(port);
		if(bind(sock, (struct sockaddr *) &sain, sizeof(sain)) >= 0)
			break;
	}
	printf("Using port %d for NFS.\n", port);
	if(port <= IPPORT_RESERVED/2){
		perror("bind to reserved port"); exit(1);
    }
	if((nfsxprt = svcudp_create(sock)) == 0){
		perror("svcudp_create"); exit(1);
	}
	if(!svc_register(nfsxprt, NFS_PROGRAM, NFS_VERSION, proc, 0)){
		perror("svc_register"); exit(1);
	}

/*** Second part: mount the directory */

	nfs_args.wsize    = 32768;
	nfs_args.rsize    = 32768;
	nfs_args.addr     = &sain;
	nfs_args.fh       = (char *)root_fh;
	nfs_args.retrans = 0;
	nfs_args.timeo = 50;	/* 5s timeout */
	nfs_args.hostname = "vmount";
	nfs_args.netname = "net";
	nfs_args.flags    = NFSMNT_SOFT | NFSMNT_HOSTNAME |
						NFSMNT_RETRANS | NFSMNT_TIMEO |
						NFSMNT_WSIZE | NFSMNT_RSIZE;

	mntdir = dir;

	my_pid = getpid();
	switch(pid = fork()){
	default:
		break;
	case -1:
		perror("fork");
		exit(1);
	case 0:
		DPRINTF(("Going to mount...\n"));
		if(mount(MOUNT_NFS, dir, wr_enable ? 0:M_RDONLY, (caddr_t)&nfs_args)){
			extern int errno;
			fprintf(stderr, "nfs mount %s: (errno %d) ", dir, errno);
			perror("");
			kill(my_pid, SIGTERM);
		}else{
			FILE *fp;
			struct mntent mnt;

			mnt.mnt_fsname = "vmount";
			mnt.mnt_dir = mntdir;
			mnt.mnt_type = "nfs";
			mnt.mnt_opts = wr_enable ? "soft" : "soft,ro";
			mnt.mnt_freq = mnt.mnt_passno = 0;
			if ((fp = setmntent(MOUNTED, "a")))
				addmntent(fp, &mnt);
			else
				perror(MOUNTED);
			endmntent(fp);
		}
		exit(0);
	}

	for(;;){
		fd_set			readfd;
		struct timeval	tv;

		readfd = svc_fdset;
		tv.tv_sec = 1; tv.tv_usec = 0;
	
		DDPRINTF(("- going into select\n"));
		if(select(FD_SETSIZE, &readfd, NULL, NULL, &tv) > 0){
			DDPRINTF(("- executing request\n"));
			svc_getreqset(&readfd);
		}
		if(do_exit){
			DDPRINTF(("- doing exit\n"));
			if(do_nfsumount)
				nfs_unmount();
			if(wr_enable){
				fprintf(stderr, "Going to unmount disk...\n");
				my_unmount();
				fprintf(stderr, "Disk unmounted.\n");
			}
			fprintf(stderr, "Exiting.\n");
			exit(0);
		}
		DDPRINTF(("- checking for regular jobs()\n"));
		dummies_regular();	/* calls all other regulars */
	}
}
